/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
 
package au.com.terabit.jproactor;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;

/**
 * Th <code>OpAccept</code> class implements <i>accept</i> IO operation.
 * <p>
 * Copyright &copy; 2003 Terabit Pty Ltd. All rights reserved.
 *
 * @author <a href="mailto:libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class OpAccept extends IOOperation
{
    
    /** Newly accepted connection instance */
    private SocketChannel m_acceptedChannel = null;

    /**
     * Creates accept operation using specified <code>AsynchChannel</code>.  
     * 
     * @param achannel <code>AsynchChannel</code> instance for this accept operation
     */
    OpAccept( AsynchChannel achannel )
    {
        super( IOOperation.OP_ACCEPT, achannel );
        this.m_state = IOOperation.STATE_STARTED;
    }

    /**
     * Accept operation on the associated server socket is ready for
     * non-blocking execution. 
     * 
     * @see IOOperation#execute()
     */
    public void execute() throws Exception
    {
        //System.out.println( Thread.currentThread().getName() + ": Executing Accept IO" );

        // get server socket channel from m_asynchChannel
        ServerSocketChannel server = (ServerSocketChannel) m_asynchChannel.getChannel();
        m_acceptedChannel = server.accept();
            
        // check that accept actually has happened 
        if ( m_acceptedChannel != null )
        {
            m_acceptedChannel.socket().setReuseAddress(true);
                
            // this will indicated that new connection accepted
            m_state = IOOperation.STATE_FINISHED;
             
            //System.out.println( Thread.currentThread().getName() +
            //                   ": Accepted new connection: " + m_acceptedChannel );
        }
    }
    
    /**
     * @see IOOperation#nextPossible()
     */
    public boolean nextPossible ()
    {
        return this.m_acceptedChannel != null;
    }

    /**
     * Dispatches accepted channel to the associated protocol so it can start
     * execution.
     * 
     * @see IOOperation#onComplete()
     * @see AsynchHandler#acceptCompleted(OpAccept)
     * @see 
     */
    public void onComplete()
    {
	try
        {
            if ( m_acceptedChannel == null )
            {
                System.out
                        .println( "OpAccept::onComplete(): no accepted channel" );
                return;
            }

            m_asynchChannel.getHandler().acceptCompleted( this );
        }
        catch ( Exception e )
        {
            System.out.println( "OpAccept::onComplete:" + e.getMessage() );
        }
    }
    
    /**
     * Returns the result of the accept operation.
     * 
     * @return a newly accepted socket channel
     */
    public SocketChannel getAcceptedChannel() { return m_acceptedChannel; }

}
