/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.jproactor;
import java.net.SocketAddress;
import java.nio.channels.SocketChannel;

/**
 * The <tt>connect</tt> operation implementation.
 *
 * @author <a href="mailto:libman@terabit.comg">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class OpConnect extends IOOperation
{  
	private SocketAddress m_remoteAddr = null;
	/* ---------------------------------------------------------------------- */
    
	OpConnect (AsynchChannel handler, SocketAddress addr) throws Exception
	{
		super( IOOperation.OP_CONNECT, handler);
		this.m_remoteAddr = addr;
		
		SocketChannel channel = (SocketChannel) handler.getChannel();
		channel.configureBlocking(false);
		
		if (channel.connect(this.m_remoteAddr))
		{
			m_state = IOOperation.STATE_FINISHED;
            m_asynchChannel.setState( AsynchChannel.CONNECTED );
		}
		else
		{
			m_state = IOOperation.STATE_STARTED;
		}
	}
    
	/* (non-Javadoc)
	 * @see IOOperation#execute()
	 */
	public void execute() throws Exception
	{
		//System.out.println( Thread.currentThread().getName() + ": Executing Read IO" );

		SocketChannel channel = (SocketChannel) m_asynchChannel.getChannel();
        
		if (channel.finishConnect())
		{
			m_state = IOOperation.STATE_FINISHED;
            m_asynchChannel.setState( AsynchChannel.CONNECTED );
		}
	}

	public boolean nextPossible ()
	{
		return true;
	}
    
	public void onComplete() 
	{
		try
		{
			m_asynchChannel.getHandler().connectCompleted( this );
		}
		catch (Exception e)
		{
			System.out.println ("OpConnect::onComplete: " +
								e.getMessage());
		}

	}

    /**
     * Tells whether connect operation on the channel has successfully completed
     * or not.
     * 
     * @return <tt>true</tt> if associated asynchronous chanel is connected,
     *         <tt>false</tt> otherwise 
     */
    public boolean isConnected()
    {
        return m_state == STATE_FINISHED && getError() == null;
    }
}
