/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.testmpx;

import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import EDU.oswego.cs.dl.util.concurrent.CountDown;
import java.text.MessageFormat;
//import java.nio.CharBuffer;

import  au.com.terabit.jproactor.*;


/**
 * A simple client protocol implementation.
 * <p>
 * This client simple connects to the server, i.e. <i>Echo</i> server,
 * sends data and then requests read operation.
 *
 * @see au.com.terabit.echoserver.EchoProtocol
 * 
 * @author <a href="mailto:libman@terabit.com.au">Alexander Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class AutoClientProtocol implements AsynchHandler 
{
	/** Default buffer size - 16 KB */
	private final static int BUFFER_SIZE = 4096;  //*4;  // 1 << 14;
    
	ByteBuffer buffer1 = ByteBuffer.allocate( BUFFER_SIZE );
    
    SocketAddress     address = null;
	AsynchChannel         achannel = null;
	AutoClientProtocolFactory factory = null;
    
    CountDown latch = null;
	
	int  traceLevel = 0;
    
	AutoClientProtocol( Demultiplexor m,  
			            SocketAddress addr ,
			            AutoClientProtocolFactory f) throws Exception
	{
		this (m,f);
		this.address = addr;
	}

	AutoClientProtocol( Demultiplexor m,
						AutoClientProtocolFactory f) throws Exception
			
	{
		this.factory = f;
		this.achannel = new AsynchChannel( m, this, SocketChannel.open() );
	}
	
	void setTraceLevel (int level)
	{
		traceLevel = level;
	}
	
	public void start(SocketAddress addr, int runtime) throws Exception
	{
        if ( runtime > 0 ) achannel.startTimer(runtime*1000);

        achannel.connect(addr);
		factory.onProtocolStarted ();
	}

	//
	// Interface AsynchHanlder
	//
    public void start() throws Exception
    { 
        start (this.address,0);
    }
    
    public void start(int runtime, CountDown latch) throws Exception
    {
        this.latch = latch;
        start (this.address,runtime);
    }

    public void channelAttached(AsynchChannel channel) throws Exception {}

    public void channelClosed(AsynchChannel channel) throws Exception
    {
    	IOStatistic rdInfo = achannel.m_readInfo;
    	IOStatistic wrInfo = achannel.m_writeInfo;
    	
    	System.out.println(
    		MessageFormat.format(
            	"Auto Client Connection Closed : Reads={0}({1},{2},{3}) Writes={4}({5},{6},{7})\n",
            	 new Object[] {
                     new Long( rdInfo.m_xferBytes ),
                     new Long( rdInfo.m_opStarted ),
                     new Long( rdInfo.m_opFinished ),
                     new Long( rdInfo.m_opCancelled ),
                     new Long( wrInfo.m_xferBytes ),
                     new Long( wrInfo.m_opStarted ),
                     new Long( wrInfo.m_opFinished ),
                     new Long( wrInfo.m_opCancelled )
                }
    		) );
            
        factory.onProtocolFinished(rdInfo, wrInfo);
        
        if ( latch != null ) latch.release();
    }

	public void acceptCompleted( OpAccept op ) throws Exception
	{
		throw new Exception ("Client::acceptCompleted - illegal call");
	}
	
	
	public void connectCompleted (OpConnect opConnect ) throws Exception
	{
		System.out.print( Thread.currentThread().getName() +
						  " Client::connectCompleted: " ); 
		 
		if ( opConnect.getError() != null )
		{    
			System.out.println( opConnect.getError().toString());
            achannel.close();
			return;
		}
		else
		{
			System.out.println( "OK");
			buffer1.clear();
			
			String demoStr = "123456789ABCDEF\n";
			
			byte [] array = demoStr.getBytes();
			
			buffer1.put(array);
			buffer1.position(0);
			buffer1.limit(1024);
			
			//buffer1.position (chBuf.position()*2);
			//buffer1.flip();
			
			achannel.write_n(buffer1);
		}
            
	}
 
	/* (non-Javadoc)
	 * @see AsynchChannel#onReadFinished(OpRead)
	 */
	public void readCompleted( OpRead opRead ) throws Exception
	{
		//System.out.println("Read finished: " + total_op_r);
        
		if ( opRead.getError() != null )
		{    
			 //System.out.println( "Client::readCompleted: " + 
			 //				  opRead.getError().toString());
            
           
			 achannel.close();
			 return;
		}
        
		if ( opRead.getBytesCompleted () <= 0)
		{
			 //System.out.println( "Client::readCompleted: Peer closed " +
			 //					opRead.getBytesCompleted() );
			 achannel.close();
			 return;
		}
        
		
		ByteBuffer buffer = opRead.getBuffer();
		
		//buffer.flip();
		
		if ( traceLevel > 0)
		{
			byte [] array = buffer.array();
		
		
			System.out.print("Recv " 
							   + opRead.getBytesCompleted ()
							   + " bytes:" );

			System.out.write(array,0,opRead.getBytesCompleted ());
		}
		
		achannel.write_n (buffer);
	}

	/* (non-Javadoc)
	 * @see 
	 */
	public void writeCompleted(OpWrite opWrite) throws Exception
	{
		
		//System.out.println("Write finished: " + total_op_w);
        
		if ( opWrite.getError() != null )
		{    
			System.out.println( "Client::writeCompleted: " + 
							  opWrite.getError().toString());
           
			achannel.close();
            return;
		}
        
		if ( opWrite.getBytesCompleted () <= 0)
		{
			 System.out.println( "Client::writeCompleted: Peer closed " +
								opWrite.getBytesCompleted() );
			 achannel.close();
			 return;
		}
        
	
        
		ByteBuffer buffer = opWrite.getBuffer();

		if ( traceLevel > 0)
		{
			buffer.flip();
		
			byte [] array = buffer.array();
		
			System.out.print("Sent " 
							   + opWrite.getBytesCompleted ()
							   + " bytes:" );
						   
			System.out.write(array,0,opWrite.getBytesCompleted ());
		
		}
		
		buffer.clear();
		achannel.read(buffer);
	}

    public void timerExpired(OpTimer opTimer) throws Exception
    {
        // stop protocol
        achannel.close();
    }

}
