/*
 *********************************************************************
 * Created on 08.02.2004
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 **********************************************************************
 */
package au.com.terabit.testmpx;

import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.nio.CharBuffer;
import java.text.MessageFormat;

import  au.com.terabit.jproactor.*;

/**
 * A Debug version of simple client protocol.
 * <p>
 * All IO operations are controlled by <code>TestClient</code>.
 *
 * @see au.com.terabit.echoserver.EchoProtocol
 * 
 * @author <a href="mailto:libman@terabit.com.au">Alexander Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class ManualClientProtocol implements AsynchHandler
{
	/** Default buffer size - 16 KB */
	private final static int BUFFER_SIZE = 4096;  //*4;  // 1 << 14;
    
	ByteBuffer rdBuffer = ByteBuffer.allocate( BUFFER_SIZE );
    
	SocketAddress     address = null;
	AsynchChannel     achannel = null; 
    

	int  traceLevel = 0;
	
    
	ManualClientProtocol( Demultiplexor m,  SocketAddress addr ) throws Exception
	{
		this (m);
		this.address = addr;
	}

	ManualClientProtocol( Demultiplexor m ) throws Exception
	{
		this.achannel = new AsynchChannel( m, this, SocketChannel.open() );
	}
	
	void setTraceLevel (int level)
	{
		traceLevel = level;
	}
	
	public void start(SocketAddress addr) throws Exception
	{
		achannel.connect( addr);
	}

	//
	// Interface ProtocolHandler
	//
	public void start() throws Exception
	{
		start (this.address);
	}
	
    public void channelAttached( AsynchChannel channel ) throws Exception {}

	public void channelClosed( AsynchChannel channel ) throws Exception
	{
    	IOStatistic rdInfo = achannel.m_readInfo;
    	IOStatistic wrInfo = achannel.m_writeInfo;
    	
        System.out.println(
            MessageFormat.format(
                "Manual Client Connection Closed : Reads={0}({1},{2},{3}) Writes={4}({5},{6},{7})\n",
                 new Object[] {
                     new Long( rdInfo.m_xferBytes ),
                     new Long( rdInfo.m_opStarted ),
                     new Long( rdInfo.m_opFinished ),
                     new Long( rdInfo.m_opCancelled ),
                     new Long( wrInfo.m_xferBytes ),
                     new Long( wrInfo.m_opStarted ),
                     new Long( wrInfo.m_opFinished ),
                     new Long( wrInfo.m_opCancelled )
                }
            ) );
	}

	public void acceptCompleted( OpAccept op ) throws Exception
	{
		throw new Exception ("Client::acceptCompleted - illegal call");
	}
	
	
	public void connectCompleted (OpConnect opConnect ) throws Exception
	{
		System.out.print( Thread.currentThread().getName() +
						  " Client::connectCompleted: " ); 
		 
		if ( opConnect.getError() != null )
		{    
			System.out.println( opConnect.getError().toString());
			achannel.close();
			return;
		}
		else
		{
			System.out.println( "OK");
			
			rdBuffer.clear();
			achannel.read (rdBuffer);
		}
            
	}
 
	/* (non-Javadoc)
	 * @see IOHandler#onReadFinished(OpRead)
	 */
	public void readCompleted( OpRead opRead ) throws Exception
	{

		//System.out.println("Read finished: " + total_op_r);
        
		if ( opRead.getError() != null )
		{    
			 System.out.println( "Client::readCompleted: " + 
							  opRead.getError().toString());
            
           
			 achannel.close();
			 return;
		}
        
		if ( opRead.getBytesCompleted () <= 0)
		{
			 System.out.println( "Client::readCompleted: Peer closed " +
								opRead.getBytesCompleted() );
			 achannel.close();
			 return;
		}
        
        
		ByteBuffer buffer = opRead.getBuffer();
		
		//buffer.flip();
		
		byte [] array = buffer.array();
		
		
		System.out.print("Recv " 
						   + opRead.getBytesCompleted ()
						   + " bytes:" );

		System.out.write(array,0,opRead.getBytesCompleted ());
								 

		// start read again						 
		buffer.clear();						
		achannel.read(buffer);
	}

	/* (non-Javadoc)
	 * @see IOHandler#onWriteFinished(OpWrite)
	 */
	public void writeCompleted(OpWrite opWrite) throws Exception
	{
		//System.out.println("Write finished: " + total_op_w);
        
		if ( opWrite.getError() != null )
		{    
			System.out.println( "Client::writeCompleted: " + 
							  opWrite.getError().toString());
           
			achannel.close();
			return;
		}
        
		if ( opWrite.getBytesCompleted () <= 0)
		{
			 System.out.println( "Client::writeCompleted: Peer closed " +
								opWrite.getBytesCompleted() );
			 achannel.close();
			 return;
		}
        
	
        
		ByteBuffer buffer = opWrite.getBuffer();

		buffer.flip();
		
		byte [] array = buffer.array();
		
	    System.out.print("Sent " 
						   + opWrite.getBytesCompleted ()
						   + " bytes:" );
						   
		System.out.write(array,0,opWrite.getBytesCompleted ());
		
		buffer = null;
	}
	
	public void write(byte[] data, int off, int len) throws Exception
	{
		if (data == null || len == 0 )
			return;
			
		ByteBuffer buffer = ByteBuffer.allocate(len);
		
		buffer.put(data, off, len);
		buffer.flip();
		
		achannel.write_n(buffer);
	}
	
	public void write(String str) throws Exception
	{
		if (str == null || str.length() == 0)
			return;
			
		ByteBuffer buffer = ByteBuffer.allocate(str.length()*2);
		
		CharBuffer chBuf = buffer.asCharBuffer();
		chBuf.put(str);
		
		achannel.write_n(buffer);
	}
	
	public void write(char ch) throws Exception
	{
			
		ByteBuffer buffer = ByteBuffer.allocate(2);
		
		buffer.putChar(ch);
		buffer.flip();
		
		achannel.write_n(buffer);
	}

    public void timerExpired(OpTimer opTimer) throws Exception {}

}
