/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.testmpx;
 
import au.com.terabit.jproactor.*;

import java.net.InetSocketAddress;
import java.text.MessageFormat;

import EDU.oswego.cs.dl.util.concurrent.CountDown;

/**
 * 
 * A simple <code>TestServer</code> that starts <code>EchoServerProtocol<code>.
 * <p>
 * This class is an example of how to use <i>Java Proator</i> classes to
 * start <code>ProtocolHandler</code> instance.
 * <p>
 * Usage:
 * <p>au.com.terabit.echoserver.TestServer [listen port number] [thread pool size]
 * Copyright &copy; 2003 Terabit Pty Ltd. All rights reserved.
 *
 * @author <a href="mailto:libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class TestMpx 
{

    /**
     * 
     */
	int m_listenPort = 12345;
	String m_listenHost = new String ("localhost");
	int m_connectPort = 12345;
	String m_connectHost = new String ("localhost");

    /** denotes whether test shall start EchoServer */ 
    boolean m_serverMode   = true;
    boolean m_autoClient   = true;
    boolean m_manualClient = false;
	
	int m_threads  = 1;
	int m_sessions = 1;
	int m_timeRun  = 0; // in seconds
    
    CountDown m_latch = null;

    long m_actualRunTime = 0;
	
	Demultiplexor m_demultiplexor = null;
	EchoServerProtocolFactory m_srvFactory = new EchoServerProtocolFactory();
	AutoClientProtocolFactory m_clnFactory = new AutoClientProtocolFactory();
	
    public TestMpx() throws Exception 
    {
    	m_demultiplexor = new Demultiplexor();
    }
    
    public void startAcceptor () throws Exception
    {
        InetSocketAddress addr = 
        	new InetSocketAddress ( m_listenHost, m_listenPort );
        
        System.out.println( "Start Acceptor on "  + addr );
        Acceptor a = new Acceptor( m_demultiplexor, m_srvFactory , addr );
        a.start(10); //  start 10 pending accepts in queue
        
        if ( m_timeRun > 0 )
        {
            AsynchHandler timerHandler = new AsynchHandler()
            {
                public void timerExpired(OpTimer opTimer) throws Exception
                {
                    if ( m_latch != null ) m_latch.release();
                }
                public void channelClosed(AsynchChannel channel) throws Exception{}
                public void channelAttached(AsynchChannel channel) throws Exception{}
                public void connectCompleted(OpConnect opConnect) throws Exception{}
                public void acceptCompleted(OpAccept opAccept) throws Exception{}
                public void writeCompleted(OpWrite opWrite) throws Exception{}
                public void readCompleted(OpRead opRead) throws Exception{}
                public void start() throws Exception{}
            };
            OpTimer timer = new OpTimer(timerHandler, m_timeRun*1000);
            m_demultiplexor.startTimer(timer);
        }
    }
    
    public void startAutoClients () throws Exception
    {
        InetSocketAddress addr = 
        	new InetSocketAddress ( m_connectHost, m_connectPort );
        
        System.out.println( "Number Auto Sessions is " + m_sessions );
        System.out.println( "Address to connect " + addr );

        for ( int i = 0; i < m_sessions; ++i )
        {
            AutoClientProtocol client = 
            	new AutoClientProtocol(m_demultiplexor, addr, m_clnFactory);

            client.setTraceLevel( 0 ); //  set 1 to see trace in console
            client.start(m_timeRun, m_latch);
        }
    }

    public void startManualClient() throws Exception
    {
        InetSocketAddress addr = 
            new InetSocketAddress ( m_connectHost, m_connectPort );

        ManualClientProtocol client =
            new ManualClientProtocol( m_demultiplexor, addr );

        client.setTraceLevel( 0 ); //  set 1 to see trace in console
        client.start();

        System.out.println( "Manual Session" );
        System.out.println( "Type any data to send or 'q' to quit==>" );

        byte[] inBuf = new byte[1024];

        for ( ;; )
        {
            int num = System.in.read( inBuf, 0, inBuf.length );

            if ( num <= 0 )          break;
            if ( inBuf[0] == 'q' )   break;

            client.write( inBuf, 0, num );
        }
    }

    public void run () throws Exception
    {
        int latchCount = 0;
        
        if ( m_serverMode ) latchCount++;
        if ( m_autoClient ) latchCount += m_sessions;
    	
        m_latch = new CountDown(latchCount);
        
        System.out.println( "Demultiplexor thread pool size is " + m_threads );
    	m_demultiplexor.start(m_threads);

        long startTime = System.currentTimeMillis();

        if ( m_serverMode )         startAcceptor();
        if ( m_autoClient )         startAutoClients ();
        
        if ( m_manualClient ) 
        {
        	startManualClient();
        }
        else
        {
            if ( m_timeRun == 0 )
            {
                System.out.println( "Press any key to stop ==>" );
                System.in.read();
            }
            else
            {
                m_latch.acquire();
            }
        }
    	m_demultiplexor.shutdown();
    	long endTime = System.currentTimeMillis();
    	
    	m_actualRunTime =  endTime - startTime;
    	printStatistics();
        System.out.println( "TestMpx finished" );
    }
    
    public void printStatistics ()
    {
        m_clnFactory.printStats();
        m_srvFactory.printStats();
        
        long rBytes = m_clnFactory.m_readInfo.m_xferBytes 
        			+ m_srvFactory.m_readInfo.m_xferBytes ;
        long rOps   = m_clnFactory.m_readInfo.m_opFinished 
					+ m_srvFactory.m_readInfo.m_opFinished ;

        long wBytes = m_clnFactory.m_writeInfo.m_xferBytes 
	    			+ m_srvFactory.m_writeInfo.m_xferBytes ;
        
        long wOps   = m_clnFactory.m_writeInfo.m_opFinished 
					+ m_srvFactory.m_writeInfo.m_opFinished ;
        
        System.out.println(
        	MessageFormat.format(
                "Total : Reads={0,number}({1}) Writes={2}({3})\n",
        		new Object[] {
                    new Long( rBytes ),
                    new Long( rOps ),
                    new Long( wBytes ),
                    new Long( wOps )
                } ) );
        
        double time = m_actualRunTime/1000;
        double avgBytes = (rBytes + wBytes)/time;
        double avgOps   = (rOps   + wOps)/time;
        
        System.out.println(
            MessageFormat.format(
                "RunTime = {0} Threads={1}\n",
				new Object[] {
                    new Double( time),
                    new Long( m_threads )
                } ) );
        
        System.out.println(
            MessageFormat.format(
                "Throughtput Bytes/sec={0} Op/sec={1}\n",
                new Object[] {
                    new Double( avgBytes ),
                    new Double( avgOps )
                } ) );
    }
    
    public void parseArgs ( String[] argv ) throws Exception
    {
    	int argc = argv.length;
    	
    	for (int i=0; i < argc-1; i+=2)
    	{
    		if (argv[i].compareTo ( "--mode") == 0 ||
           			argv[i].compareTo ( "-m" ) == 0 )
    		{
    			 m_serverMode = false;
    			 m_autoClient = false;
    			 m_manualClient = false;
    			for (int j = 0; j < argv[i+1].length(); ++j)
    			{
    				char c = Character.toLowerCase( argv[i+1].charAt(j));
    				switch (c)
    				{
    				case 's':  m_serverMode   = true; break;
    				case 'a':  m_autoClient   = true; break;
    				case 'm':  m_manualClient = true; break;
    				default:   System.out.println("Invalid paramter mode\n");
    				}
    			}
    		}
            else if (argv[i].compareTo ( "--threads") == 0 ||
       			argv[i].compareTo ( "-t" ) == 0 )
       		{
    			m_threads =	Integer.parseInt( argv[i+1] );
       		}
       		else if (argv[i].compareTo ( "--sessions") == 0 || 
       				 argv[i].compareTo ( "-s" ) == 0 )
       		{
       			m_sessions = Integer.parseInt( argv[i+1] );
                
       		}
    		else if (argv[i].compareTo ( "--port_listen") == 0 ||
    				 argv[i].compareTo ( "-pl" ) == 0 )
    		{
    			m_listenPort = Integer.parseInt( argv[i+1] );
    		}
    		else if (argv[i].compareTo ( "--port_connect") == 0 ||
    				 argv[i].compareTo ( "-pc" ) == 0 )
    		{
    			m_connectPort = Integer.parseInt( argv[i+1] );
    		}
    		else if (argv[i].compareTo ( "--host_listen") == 0 || 
    				 argv[i].compareTo ( "-hl" ) == 0 )
    		{
    			m_listenHost = argv[i+1];
    		}
    		else if (argv[i].compareTo ( "--host_connect") == 0 || 
      				 argv[i].compareTo ( "-hc" ) == 0 )
    		{
    			m_connectHost = argv[i+1];
    		}
    		else if (argv[i].compareTo ( "--time") ==0  ||
    				 argv[i].compareTo ( "-i" ) == 0)
    		{
    			m_timeRun  = Integer.parseInt( argv[i+1] );
    		}
    		else 
    		{
    			usage ();
    		}
    	}
    }
    
	public static void main( String[] argv ) throws Exception
    {
    	TestMpx test = new TestMpx ();
    	test.parseArgs (argv);
    	test.run ();
    }
    
    private static void usage()
    {
        System.out.println( "Usage: " + TestMpx.class.getName() + 
        		            "\n [--mode         [S][A][M]]" +
                            "\n [--threads      number]" +
                            "\n [--sessions     number]" +
                            "\n [--time         seconds]" +
                            "\n [--port_listen  port]" +
                            "\n [--host_listen  host]" +
                            "\n [--port_connect port]" +
                            "\n [--host_connect host]" +
                            "\n"
                          );
        System.exit( -1 );
    }

}
